@@ -19,13 +19,14 @@
 #include <QtWaylandClientVersion>
 
 #include <errno.h>
+#include <fcntl.h>
 #include <poll.h>
 #include <signal.h>
 #include <string.h>
 #include <unistd.h>
 
 #include "qwayland-wayland.h"
-#include "qwayland-wlr-data-control-unstable-v1.h"
+#include "qwayland-ext-data-control-v1.h"
 
 static inline QString applicationQtXImageLiteral()
 {
@@ -62,12 +63,12 @@ static inline QStringList imageWriteMimeFormats()
 }
 // end copied
 
-class DataControlDeviceManager : public QWaylandClientExtensionTemplate<DataControlDeviceManager>, public QtWayland::zwlr_data_control_manager_v1
+class DataControlDeviceManager : public QWaylandClientExtensionTemplate<DataControlDeviceManager>, public QtWayland::ext_data_control_manager_v1
 {
     Q_OBJECT
 public:
     DataControlDeviceManager()
-        : QWaylandClientExtensionTemplate<DataControlDeviceManager>(2)
+        : QWaylandClientExtensionTemplate<DataControlDeviceManager>(1)
     {
     }
 
@@ -84,12 +85,12 @@ public:
     }
 };
 
-class DataControlOffer : public QMimeData, public QtWayland::zwlr_data_control_offer_v1
+class DataControlOffer : public QMimeData, public QtWayland::ext_data_control_offer_v1
 {
     Q_OBJECT
 public:
-    DataControlOffer(struct ::zwlr_data_control_offer_v1 *id)
-        : QtWayland::zwlr_data_control_offer_v1(id)
+    DataControlOffer(struct ::ext_data_control_offer_v1 *id)
+        : QtWayland::ext_data_control_offer_v1(id)
     {
     }
 
@@ -144,7 +145,7 @@ public:
     }
 
 protected:
-    void zwlr_data_control_offer_v1_offer(const QString &mime_type) override
+    void ext_data_control_offer_v1_offer(const QString &mime_type) override
     {
         if (!m_receivedFormats.contains(mime_type)) {
             m_receivedFormats << mime_type;
@@ -234,11 +235,7 @@ QVariant DataControlOffer::retrieveData(const QString &mimeType, QMetaType type)
                 QVariantList list;
                 list.reserve(urls.size());
                 for (const QByteArray &s : urls) {
-#if QT_VERSION >= QT_VERSION_CHECK(6, 7, 0)
                     if (QUrl url(QUrl::fromEncoded(QByteArrayView(s).trimmed())); url.isValid()) {
-#else
-                    if (QUrl url(QUrl::fromEncoded(QByteArrayView(s).trimmed().toByteArray())); url.isValid()) {
-#endif
                         list.emplace_back(std::move(url));
                     }
                 }
@@ -286,11 +283,11 @@ bool DataControlOffer::readData(int fd, QByteArray &data)
     }
 }
 
-class DataControlSource : public QObject, public QtWayland::zwlr_data_control_source_v1
+class DataControlSource : public QObject, public QtWayland::ext_data_control_source_v1
 {
     Q_OBJECT
 public:
-    DataControlSource(struct ::zwlr_data_control_source_v1 *id, QMimeData *mimeData);
+    DataControlSource(struct ::ext_data_control_source_v1 *id, QMimeData *mimeData);
     DataControlSource() = default;
     ~DataControlSource()
     {
@@ -310,15 +307,15 @@ Q_SIGNALS:
     void cancelled();
 
 protected:
-    void zwlr_data_control_source_v1_send(const QString &mime_type, int32_t fd) override;
-    void zwlr_data_control_source_v1_cancelled() override;
+    void ext_data_control_source_v1_send(const QString &mime_type, int32_t fd) override;
+    void ext_data_control_source_v1_cancelled() override;
 
 private:
     std::unique_ptr<QMimeData> m_mimeData;
 };
 
-DataControlSource::DataControlSource(struct ::zwlr_data_control_source_v1 *id, QMimeData *mimeData)
-    : QtWayland::zwlr_data_control_source_v1(id)
+DataControlSource::DataControlSource(struct ::ext_data_control_source_v1 *id, QMimeData *mimeData)
+    : QtWayland::ext_data_control_source_v1(id)
     , m_mimeData(mimeData)
 {
     const auto formats = mimeData->formats();
@@ -340,7 +337,7 @@ DataControlSource::DataControlSource(struct ::zwlr_data_control_source_v1 *id, Q
     }
 }
 
-void DataControlSource::zwlr_data_control_source_v1_send(const QString &mime_type, int32_t fd)
+void DataControlSource::ext_data_control_source_v1_send(const QString &mime_type, int32_t fd)
 {
     QString send_mime_type = mime_type;
     if (send_mime_type == QStringLiteral("text/plain;charset=utf-8")) {
@@ -369,6 +366,10 @@ void DataControlSource::zwlr_data_control_source_v1_send(const QString &mime_typ
         ba = m_mimeData->data(send_mime_type);
     }
 
+    QFile c;
+    if (!c.open(fd, QFile::WriteOnly, QFile::AutoCloseHandle)) {
+        return;
+    }
     // Create a sigpipe handler that does nothing, or clients may be forced to terminate
     // if the pipe is closed in the other end.
     struct sigaction action, oldAction;
@@ -376,22 +377,29 @@ void DataControlSource::zwlr_data_control_source_v1_send(const QString &mime_typ
     sigemptyset(&action.sa_mask);
     action.sa_flags = 0;
     sigaction(SIGPIPE, &action, &oldAction);
-    write(fd, ba.constData(), ba.size());
+    const int flags = fcntl(fd, F_GETFL, 0);
+    if (flags & O_NONBLOCK) {
+        fcntl(fd, F_SETFL, flags & ~O_NONBLOCK); // Unset O_NONBLOCK to fix pasting to XWayland windows
+    }
+    const qint64 written = c.write(ba);
     sigaction(SIGPIPE, &oldAction, nullptr);
-    close(fd);
+
+    if (written != ba.size()) {
+        qWarning() << "Failed to send all clipobard data; sent" << written << "bytes out of" << ba.size();
+    }
 }
 
-void DataControlSource::zwlr_data_control_source_v1_cancelled()
+void DataControlSource::ext_data_control_source_v1_cancelled()
 {
     Q_EMIT cancelled();
 }
 
-class DataControlDevice : public QObject, public QtWayland::zwlr_data_control_device_v1
+class DataControlDevice : public QObject, public QtWayland::ext_data_control_device_v1
 {
     Q_OBJECT
 public:
-    DataControlDevice(struct ::zwlr_data_control_device_v1 *id)
-        : QtWayland::zwlr_data_control_device_v1(id)
+    DataControlDevice(struct ::ext_data_control_device_v1 *id)
+        : QtWayland::ext_data_control_device_v1(id)
     {
     }
 
@@ -428,31 +436,31 @@ Q_SIGNALS:
     void primarySelectionChanged();
 
 protected:
-    void zwlr_data_control_device_v1_data_offer(struct ::zwlr_data_control_offer_v1 *id) override
+    void ext_data_control_device_v1_data_offer(struct ::ext_data_control_offer_v1 *id) override
     {
         // this will become memory managed when we retrieve the selection event
         // a compositor calling data_offer without doing that would be a bug
         new DataControlOffer(id);
     }
 
-    void zwlr_data_control_device_v1_selection(struct ::zwlr_data_control_offer_v1 *id) override
+    void ext_data_control_device_v1_selection(struct ::ext_data_control_offer_v1 *id) override
     {
         if (!id) {
             m_receivedSelection.reset();
         } else {
-            auto derivated = QtWayland::zwlr_data_control_offer_v1::fromObject(id);
+            auto derivated = QtWayland::ext_data_control_offer_v1::fromObject(id);
             auto offer = dynamic_cast<DataControlOffer *>(derivated); // dynamic because of the dual inheritance
             m_receivedSelection.reset(offer);
         }
         Q_EMIT receivedSelectionChanged();
     }
 
-    void zwlr_data_control_device_v1_primary_selection(struct ::zwlr_data_control_offer_v1 *id) override
+    void ext_data_control_device_v1_primary_selection(struct ::ext_data_control_offer_v1 *id) override
     {
         if (!id) {
             m_receivedPrimarySelection.reset();
         } else {
-            auto derivated = QtWayland::zwlr_data_control_offer_v1::fromObject(id);
+            auto derivated = QtWayland::ext_data_control_offer_v1::fromObject(id);
             auto offer = dynamic_cast<DataControlOffer *>(derivated); // dynamic because of the dual inheritance
             m_receivedPrimarySelection.reset(offer);
         }
@@ -485,10 +493,8 @@ void DataControlDevice::setPrimarySelection(std::unique_ptr<DataControlSource> s
         m_primarySelection.reset();
     });
 
-    if (zwlr_data_control_device_v1_get_version(object()) >= ZWLR_DATA_CONTROL_DEVICE_V1_SET_PRIMARY_SELECTION_SINCE_VERSION) {
-        set_primary_selection(m_primarySelection->object());
-        Q_EMIT primarySelectionChanged();
-    }
+    set_primary_selection(m_primarySelection->object());
+    Q_EMIT primarySelectionChanged();
 }
 class Keyboard;
 // We are binding to Seat/Keyboard manually because we want to react to gaining focus but inside Qt the events are Qt and arrive to late
@@ -607,6 +613,16 @@ WaylandClipboard::WaylandClipboard(QObject *parent)
 
 WaylandClipboard::~WaylandClipboard() = default;
 
+WaylandClipboard *WaylandClipboard::create(QObject *parent)
+{
+    auto clipboard = new WaylandClipboard(parent);
+    if (clipboard->isValid()) {
+        return clipboard;
+    }
+    delete clipboard;
+    return nullptr;
+}
+
 bool WaylandClipboard::isValid()
 {
     return m_manager && m_manager->isInitialized();
@@ -626,7 +642,7 @@ void WaylandClipboard::setMimeData(QMimeData *mime, QClipboard::Mode mode)
     // If the application is focused, use the normal mechanism so a future paste will not deadlock itselfs
     if (m_keyboardFocusWatcher->hasFocus()) {
         QGuiApplication::clipboard()->setMimeData(mime, mode);
-        // if we short-circuit the wlr_data_device, when we receive the data
+        // if we short-circuit the ext_data_device, when we receive the data
         // we cannot identify ourselves as the owner
         // because of that we act like it's a synchronous action to not confuse klipper.
         wl_display_roundtrip(display);
@@ -667,10 +683,8 @@ void WaylandClipboard::clear(QClipboard::Mode mode)
         m_device->set_selection(nullptr);
         m_device->m_selection.reset();
     } else if (mode == QClipboard::Selection) {
-        if (zwlr_data_control_device_v1_get_version(m_device->object()) >= ZWLR_DATA_CONTROL_DEVICE_V1_SET_PRIMARY_SELECTION_SINCE_VERSION) {
-            m_device->set_primary_selection(nullptr);
-            m_device->m_primarySelection.reset();
-        }
+        m_device->set_primary_selection(nullptr);
+        m_device->m_primarySelection.reset();
     }
 }
 
